"""
A generic server is an abstraction of a server loop built on top of the mailbox
module.

It is best used to build components that accept request from other components in
your application such as:

 - an in-memory key-value store
 - a TCP server handler
 - a finite state machine

There are 3 ways of sending messages to a generic server:

 - **cast:** send a message
 - **call:** send a message an wait for a response
 - directly to the mailbox

> **NB:** If a call returns an exception to the caller, the exception will be
> raised on the caller side.

.. code-block:: python
   :caption: Example

   from triotp.helpers import current_module
   from triotp import gen_server, mailbox


   __module__ = current_module()


   async def start():
       await gen_server.start(__module__, name='kvstore')


   async def get(key):
       return await gen_server.call('kvstore', ('get', key))


   async def set(key, val):
       return await gen_server.call('kvstore', ('set', key, val))


   async def stop():
       await gen_server.cast('kvstore', 'stop')


   async def printstate():
       await mailbox.send('kvstore', 'printstate')

   # gen_server callbacks

   async def init(_init_arg):
       state = {}
       return state


   # optional
   async def terminate(reason, state):
       if reason is not None:
           print('An error occured:', reason)

       print('Exited with state:', state)


   # if not defined, the gen_server will stop with a NotImplementedError when
   # receiving a call
   async def handle_call(message, _caller, state):
       match message:
           case ('get', key):
               val = state.get(key, None)
               return (gen_server.Reply(payload=val), state)

           case ('set', key, val):
               prev = state.get(key, None)
               state[key] = val
               return (gen_server.Reply(payload=prev), state)

           case _:
               exc = NotImplementedError('unknown request')
               return (gen_server.Reply(payload=exc), state)


   # if not defined, the gen_server will stop with a NotImplementedError when
   # receiving a cast
   async def handle_cast(message, state):
       match message:
           case 'stop':
               return (gen_server.Stop(), state)

           case _:
               print('unknown request')
               return (gen_server.NoReply(), state)


   # optional
   async def handle_info(message, state):
       match message:
           case 'printstate':
               print(state)

           case _:
               pass

       return (gen_server.NoReply(), state)
"""

from triotp import mailbox, logging

from dataclasses import dataclass
import trio

from typing import TypeVar, Union, Optional, Any
from types import ModuleType


State = TypeVar("State")


class GenServerExited(Exception):
    """
    Raised when the generic server exited during a call.
    """


@dataclass
class _Loop:
    yes: bool


@dataclass
class _Raise:
    exc: BaseException


Continuation = Union[_Loop, _Raise]


@dataclass
class Reply:
    """
    Return an instance of this class to send a reply to the caller.
    """

    payload: Any  #: The response to send back


@dataclass
class NoReply:
    """
    Return an instance of this class to not send a reply to the caller.
    """


@dataclass
class Stop:
    """
    Return an instance of this class to stop the generic server.
    """

    reason: Optional[BaseException] = (
        None  #: Eventual exception that caused the gen_server to stop
    )


@dataclass
class _CallMessage:
    source: trio.MemorySendChannel
    payload: Any


@dataclass
class _CastMessage:
    payload: Any


async def start(
    module: ModuleType,
    init_arg: Optional[Any] = None,
    name: Optional[str] = None,
) -> None:
    """
    Starts the generic server loop.

    :param module: Module containing the generic server's callbacks
    :param init_arg: Optional argument passed to the `init` callback
    :param name: Optional name to use to register the generic server's mailbox

    :raises triotp.mailbox.NameAlreadyExist: If the `name` was already registered
    :raises Exception: If the generic server terminated with a non-null reason
    """

    await _loop(module, init_arg, name)


async def call(
    name_or_mid: Union[str, mailbox.MailboxID],
    payload: Any,
    timeout: Optional[float] = None,
) -> Any:
    """
    Send a request to the generic server and wait for a response.

    This function creates a temporary bi-directional channel. The writer is
    passed to the `handle_call` function and is used to send the response back
    to the caller.

    :param name_or_mid: The generic server's mailbox identifier
    :param payload: The message to send to the generic server
    :param timeout: Optional timeout after which this function fails
    :returns: The response from the generic server
    :raises GenServerExited: If the generic server exited after handling the call
    :raises Exception: If the response is an exception

    """

    wchan, rchan = trio.open_memory_channel(0)
    message = _CallMessage(source=wchan, payload=payload)

    await mailbox.send(name_or_mid, message)

    try:
        if timeout is not None:
            with trio.fail_after(timeout):
                val = await rchan.receive()

        else:
            val = await rchan.receive()

        if isinstance(val, Exception):
            raise val

        return val

    finally:
        await wchan.aclose()
        await rchan.aclose()


async def cast(
    name_or_mid: Union[str, mailbox.MailboxID],
    payload: Any,
) -> None:
    """
    Send a message to the generic server without expecting a response.

    :param name_or_mid: The generic server's mailbox identifier
    :param payload: The message to send
    """

    message = _CastMessage(payload=payload)
    await mailbox.send(name_or_mid, message)


async def reply(caller: trio.MemorySendChannel, response: Any) -> None:
    """
    The `handle_call` callback can start a background task to handle a slow
    request and return a `NoReply` instance. Use this function in the background
    task to send the response to the caller at a later time.

    :param caller: The caller `SendChannel` to use to send the response
    :param response: The response to send back to the caller

    .. code-block:: python
       :caption: Example

       from triotp import gen_server, supervisor, dynamic_supervisor
       import trio


       async def slow_task(message, caller):
           # do stuff with message
           await gen_server.reply(caller, response)


       async def handle_call(message, caller, state):
           await dynamic_supervisor.start_child(
               'slow-task-pool',
               supervisor.child_spec(
                   id='some-slow-task',
                   task=slow_task,
                   args=[message, caller],
                   restart=supervisor.restart_strategy.TEMPORARY,
               ),
           )

           return (gen_server.NoReply(), state)
    """

    await caller.send(response)


async def _loop(
    module: ModuleType,
    init_arg: Optional[Any],
    name: Optional[str],
) -> None:
    async with mailbox.open(name) as mid:
        try:
            state = await _init(module, init_arg)
            looping = True

            while looping:
                message = await mailbox.receive(mid)

                match message:
                    case _CallMessage(source, payload):
                        continuation, state = await _handle_call(
                            module, payload, source, state
                        )

                    case _CastMessage(payload):
                        continuation, state = await _handle_cast(module, payload, state)

                    case _:
                        continuation, state = await _handle_info(module, message, state)

                match continuation:
                    case _Loop(yes=False):
                        looping = False

                    case _Loop(yes=True):
                        looping = True

                    case _Raise(exc=err):
                        raise err

        except Exception as err:
            await _terminate(module, err, state)
            raise err from None

        else:
            await _terminate(module, None, state)


async def _init(module: ModuleType, init_arg: Any) -> State:
    return await module.init(init_arg)


async def _terminate(
    module: ModuleType,
    reason: Optional[BaseException],
    state: State,
) -> None:
    handler = getattr(module, "terminate", None)
    if handler is not None:
        await handler(reason, state)

    elif reason is not None:
        logger = logging.getLogger(module.__name__)
        logger.exception(reason)


async def _handle_call(
    module: ModuleType,
    message: Any,
    source: trio.MemorySendChannel,
    state: State,
) -> tuple[Continuation, State]:
    handler = getattr(module, "handle_call", None)
    if handler is None:
        raise NotImplementedError(f"{module.__name__}.handle_call")

    result = await handler(message, source, state)

    match result:
        case (Reply(payload), new_state):
            state = new_state
            await reply(source, payload)
            continuation = _Loop(yes=True)

        case (NoReply(), new_state):
            state = new_state
            continuation = _Loop(yes=True)

        case (Stop(reason), new_state):
            state = new_state
            await reply(source, GenServerExited())

            if reason is not None:
                continuation = _Raise(reason)

            else:
                continuation = _Loop(yes=False)

        case _:
            raise TypeError(
                f"{module.__name__}.handle_call did not return a valid value"
            )

    return continuation, state


async def _handle_cast(
    module: ModuleType,
    message: Any,
    state: State,
) -> tuple[Continuation, State]:
    handler = getattr(module, "handle_cast", None)
    if handler is None:
        raise NotImplementedError(f"{module.__name__}.handle_cast")

    result = await handler(message, state)

    match result:
        case (NoReply(), new_state):
            state = new_state
            continuation = _Loop(yes=True)

        case (Stop(reason), new_state):
            state = new_state

            if reason is not None:
                continuation = _Raise(reason)

            else:
                continuation = _Loop(yes=False)

        case _:
            raise TypeError(
                f"{module.__name__}.handle_cast did not return a valid value"
            )

    return continuation, state


async def _handle_info(
    module: ModuleType,
    message: Any,
    state: State,
) -> tuple[Continuation, State]:
    handler = getattr(module, "handle_info", None)
    if handler is None:
        return _Loop(yes=True), state

    result = await handler(message, state)

    match result:
        case (NoReply(), new_state):
            state = new_state
            continuation = _Loop(yes=True)

        case (Stop(reason), new_state):
            state = new_state

            if reason is not None:
                continuation = _Raise(reason)

            else:
                continuation = _Loop(yes=False)

        case _:
            raise TypeError(
                f"{module.__name__}.handle_info did not return a valid value"
            )

    return continuation, state



"""
A supervisor is used to handle a set of asynchronous tasks. It takes care of
restarting them if they exit prematurely or if they crash.

.. code-block:: python
   :caption: Example

   from triotp import supervisor
   from random import random
   import trio

   async def loop(threshold):
       while True:
           if random() < threshold:
               raise RuntimeError('bad luck')

           else:
               await trio.sleep(0.1)

   async def start_supervisor():
       children = [
           supervisor.child_spec(
               id='loop',
               task=loop,
               args=[0.5],
               restart=supervisor.restart_strategy.PERMANENT,
           ),
       ]
       opts = supervisor.options(
           max_restarts=3,
           max_seconds=5
       )
       await supervisor.start(children, opts)
"""

from dataclasses import dataclass
from collections import deque
from enum import Enum, auto
from logbook import Logger
import tenacity

import trio_util
import trio

from collections.abc import Callable, Awaitable
from typing import Any


class restart_strategy(Enum):
    """
    Describe when to restart an asynchronous task.
    """

    PERMANENT = auto()  #: Always restart the task
    TRANSIENT = auto()  #: Restart the task only if it raises an exception
    TEMPORARY = auto()  #: Never restart a task


@dataclass
class child_spec:
    """
    Describe an asynchronous task to supervise.
    """

    id: str  #: Task identifier
    task: Callable[..., Awaitable[None]]  #: The task to run
    args: list[Any]  #: Arguments to pass to the task
    restart: restart_strategy = restart_strategy.PERMANENT  #: When to restart the task


@dataclass
class options:
    """
    Describe the options for the supervisor.
    """

    max_restarts: int = 3  #: Maximum number of restart during a limited timespan
    max_seconds: int = 5  #: Timespan duration


class _retry_strategy:
    def __init__(
        self,
        restart: restart_strategy,
        max_restarts: int,
        max_seconds: float,
    ):
        self.restart = restart
        self.max_restarts = max_restarts
        self.max_seconds = max_seconds

        self.failure_times = deque()

    def __call__(self, retry_state: tenacity.RetryCallState):
        match self.restart:
            case restart_strategy.PERMANENT:
                pass

            case restart_strategy.TRANSIENT:
                if not retry_state.outcome.failed:
                    return False

            case restart_strategy.TEMPORARY:
                return False

        now = trio.current_time()
        self.failure_times.append(now)

        if len(self.failure_times) <= self.max_restarts:
            return True

        oldest_failure = self.failure_times.popleft()
        return now - oldest_failure >= self.max_seconds


class _retry_logger:
    def __init__(self, child_id: str):
        self.logger = Logger(child_id)

    def __call__(self, retry_state: tenacity.RetryCallState) -> None:
        if isinstance(retry_state.outcome.exception(), trio.Cancelled):
            self.logger.info("task cancelled")

        elif retry_state.outcome.failed:
            exception = retry_state.outcome.exception()
            exc_info = (exception.__class__, exception, exception.__traceback__)
            self.logger.error("restarting task after failure", exc_info=exc_info)

        else:
            self.logger.error("restarting task after unexpected exit")


async def start(
    child_specs: list[child_spec],
    opts: options,
    task_status=trio.TASK_STATUS_IGNORED,
) -> None:
    """
    Start the supervisor and its children.

    :param child_specs: Asynchronous tasks to supervise
    :param opts: Supervisor options
    :param task_status: Used to notify the trio nursery that the task is ready

    .. code-block:: python
       :caption: Example

       from triotp import supervisor
       import trio

       async def example():
           children_a = [
               # ...
           ]
           children_b = [
               # ...
           ]
           opts = supervisor.options()

           async with trio.open_nursery() as nursery:
               await nursery.start(supervisor.start, children_a, opts)
               await nursery.start(supervisor.start, children_b, opts)
    """

    async with trio.open_nursery() as nursery:
        for spec in child_specs:
            await nursery.start(_child_monitor, spec, opts)

        task_status.started(None)


async def _child_monitor(
    spec: child_spec,
    opts: options,
    task_status=trio.TASK_STATUS_IGNORED,
) -> None:
    task_status.started(None)

    @tenacity.retry(
        retry=_retry_strategy(spec.restart, opts.max_restarts, opts.max_seconds),
        reraise=True,
        sleep=trio.sleep,
        after=_retry_logger(spec.id),
    )
    async def _child_runner():
        with trio_util.defer_to_cancelled():
            async with trio.open_nursery() as nursery:
                nursery.start_soon(spec.task, *spec.args)

    await _child_runner()



"""
In Erlang_/Elixir_, each process have a PID that can be used to receive message
from other processes.

.. _erlang: https://erlang.org
.. _elixir: https://elixir-lang.org/

With trio, there is no such thing as a process. There is only asynchronous tasks
started within a nursery.

This module provides an encapsulation of trio's memory channels_ which allows
tasks to communicate with each other.

.. _channels: https://trio.readthedocs.io/en/stable/reference-core.html#using-channels-to-pass-values-between-tasks

.. code-block:: python
   :caption: Example

   from triotp import mailbox


   async def task_a(task_status=trio.TASK_STATUS_IGNORED):
       async with mailbox.open(name='task_a') as mid:
           task_status.started(None)

           msg = await mailbox.receive(mid)
           print(msg)


   async def task_b():
       await mailbox.send('task_a', 'hello world')


   async def main():
       async with trio.open_nursery() as nursery:
           await nursery.start(task_a)
           nursery.start_soon(task_b)
"""

from contextlib import asynccontextmanager
from contextvars import ContextVar
from uuid import uuid4
import trio

from collections.abc import Callable, Awaitable
from typing import TypeVar, Union, Optional, Any, AsyncContextManager


MailboxID = TypeVar("MailboxID", bound=str)  #: Mailbox identifier (UUID4)


context_mailbox_registry = ContextVar("mailbox_registry")
context_name_registry = ContextVar("name_registry")


class MailboxDoesNotExist(RuntimeError):
    """
    Error thrown when the mailbox identifier was not found.
    """

    def __init__(self, mid: MailboxID):
        super().__init__(f"mailbox {mid} does not exist")


class NameAlreadyExist(RuntimeError):
    """
    Error thrown when trying to register a mailbox to an already registered
    name.
    """

    def __init__(self, name: str):
        super().__init__(f"mailbox {name} already registered")


class NameDoesNotExist(RuntimeError):
    """
    Error thrown when trying to unregister a non-existing name.
    """

    def __init__(self, name: str):
        super().__init__(f"mailbox {name} does not exist")


def _init() -> None:
    context_mailbox_registry.set({})
    context_name_registry.set({})


def create() -> MailboxID:
    """
    Create a new mailbox.

    :returns: The mailbox unique identifier
    """

    mid = str(uuid4())

    mailbox_registry = context_mailbox_registry.get()
    mailbox_registry[mid] = trio.open_memory_channel(0)

    return mid


async def destroy(mid: MailboxID) -> None:
    """
    Close and destroy a mailbox.

    :param mid: The mailbox identifier
    :raises MailboxDoesNotExist: The mailbox identifier was not found
    """

    mailbox_registry = context_mailbox_registry.get()

    if mid not in mailbox_registry:
        raise MailboxDoesNotExist(mid)

    unregister_all(mid)

    wchan, rchan = mailbox_registry.pop(mid)
    await wchan.aclose()
    await rchan.aclose()


def register(mid: MailboxID, name: str) -> None:
    """
    Assign a name to a mailbox.

    :param mid: The mailbox identifier
    :param name: The new name

    :raises MailboxDoesNotExist: The mailbox identifier was not found
    :raises NameAlreadyExist: The name was already registered
    """

    mailbox_registry = context_mailbox_registry.get()

    if mid not in mailbox_registry:
        raise MailboxDoesNotExist(mid)

    name_registry = context_name_registry.get()
    if name in name_registry:
        raise NameAlreadyExist(name)

    name_registry[name] = mid


def unregister(name: str) -> None:
    """
    Unregister a mailbox's name.

    :param name: The name to unregister
    :raises NameDoesNotExist: The name was not found
    """

    name_registry = context_name_registry.get()
    if name not in name_registry:
        raise NameDoesNotExist(name)

    name_registry.pop(name)


def unregister_all(mid: MailboxID) -> None:
    """
    Unregister all names associated to a mailbox.

    :param mid: The mailbox identifier
    """

    name_registry = context_name_registry.get()

    for name, mailbox_id in list(name_registry.items()):
        if mailbox_id == mid:
            name_registry.pop(name)


@asynccontextmanager
async def open(name: Optional[str] = None) -> AsyncContextManager[MailboxID]:
    """
    Shortcut for `create()`, `register()` followed by a `destroy()`.

    :param name: Optional name to register the mailbox
    :returns: Asynchronous context manager for the mailbox
    :raises NameAlreadyExist: If the `name` was already registered

    .. code-block:: python
       :caption: Example

       async with mailbox.open(name='foo') as mid:
           message = await mailbox.receive()
           print(message)
    """

    mid = create()

    try:
        if name is not None:
            register(mid, name)

        yield mid

    finally:
        await destroy(mid)


def _resolve(name: str) -> Optional[MailboxID]:
    name_registry = context_name_registry.get()
    return name_registry.get(name)


async def send(name_or_mid: Union[str, MailboxID], message: Any) -> None:
    """
    Send a message to a mailbox.

    :param name_or_mid: Either a registered name, or the mailbox identifier
    :param message: The message to send
    :raises MailboxDoesNotExist: The mailbox was not found
    """

    mailbox_registry = context_mailbox_registry.get()

    mid = _resolve(name_or_mid)
    if mid is None:
        mid = name_or_mid

    if mid not in mailbox_registry:
        raise MailboxDoesNotExist(mid)

    wchan, _ = mailbox_registry.get(mid)
    await wchan.send(message)


async def receive(
    mid: MailboxID,
    timeout: Optional[float] = None,
    on_timeout: Callable[[], Awaitable[Any]] = None,
) -> Any:
    """
    Consume a message from a mailbox.

    :param mid: The mailbox identifier
    :param timeout: If set, the call will fail after the timespan set in seconds
    :param on_timeout: If set and `timeout` is set, instead of raising an
                       exception, the result of this async function will be
                       returned

    :raises MailboxDoesNotExist: The mailbox was not found
    :raises trio.TooSlowError: If `timeout` is set, but `on_timeout` isn't, and
                               no message was received during the timespan set
    """

    mailbox_registry = context_mailbox_registry.get()

    if mid not in mailbox_registry:
        raise MailboxDoesNotExist(mid)

    _, rchan = mailbox_registry.get(mid)

    if timeout is not None:
        try:
            with trio.fail_after(timeout):
                return await rchan.receive()

        except trio.TooSlowError:
            if on_timeout is None:
                raise

            return await on_timeout()

    else:
        return await rchan.receive()



"""
A TriOTP node encapsulates the call to `trio.run` and allows you to specify a
list of application to start.

   **NB:** There is no dependency management between applications, it's up to
   you to start the correct applications in the right order.

.. code-block:: python
   :caption: Example

   from pyotp import node, application

   from myproject import myapp1, myapp2

   node.run(
       apps=[
           application.app_spec(
               module=myapp1,
               start_arg=[],
           ),
           application.app_spec(
               module=myapp2,
               start_arg=[],
               permanent=False,
           ),
       ],
   )
"""

from triotp import mailbox, application, logging
from logbook import StreamHandler, NullHandler
import trio
import sys

from typing import Optional


def run(
    apps: list[application.app_spec],
    loglevel: logging.LogLevel = logging.LogLevel.NONE,
    logformat: Optional[str] = None,
) -> None:
    """
    Start a new node by calling `trio.run`.

    :param apps: List of application to start
    :param loglevel: Logging Level of the node
    :param logformat: Format of log messages produced by the node
    """

    match loglevel:
        case logging.LogLevel.NONE:
            handler = NullHandler()

        case _:
            handler = StreamHandler(sys.stdout, level=loglevel.to_logbook())

    if logformat is not None:
        handler.format_string = logformat

    with handler.applicationbound():
        trio.run(_start, apps)


async def _start(apps: list[application.app_spec]) -> None:
    mailbox._init()

    async with trio.open_nursery() as nursery:
        application._init(nursery)

        for app_spec in apps:
            await application.start(app_spec)



"""
An application is a Python module defining an asynchronous function `start`.

.. code-block:: python
   :caption: Example

   async def start(_start_arg):
       print('Hello world')

Usually, the application will start a supervisor containing the child tasks to
run.
"""

from triotp import supervisor

from contextvars import ContextVar
from dataclasses import dataclass
import trio

from typing import Optional, Any
from types import ModuleType


context_app_nursery = ContextVar("app_nursery")
context_app_registry = ContextVar("app_registry")


@dataclass
class app_spec:
    """Describe an application"""

    module: ModuleType  #: Application module
    start_arg: Any  #: Argument to pass to the module's start function
    permanent: bool = (
        True  #: If `False`, the application won't be restarted if it exits
    )
    opts: Optional[supervisor.options] = (
        None  #: Options for the supervisor managing the application task
    )


def _init(nursery: trio.Nursery) -> None:
    context_app_nursery.set(nursery)
    context_app_registry.set({})


async def start(app: app_spec) -> None:
    """
    Starts an application on the current node. If the application is already
    started, it does nothing.

       **NB:** This function cannot be called outside a node.

    :param app: The application to start
    """

    nursery = context_app_nursery.get()
    registry = context_app_registry.get()

    if app.module.__name__ not in registry:
        local_nursery = await nursery.start(_app_scope, app)
        registry[app.module.__name__] = local_nursery


async def stop(app_name: str) -> None:
    """
    Stops an application. If the application was not running, it does nothing.

       **NB:** This function cannot be called outside a node.

    :param app_name: `__name__` of the application module
    """

    registry = context_app_registry.get()

    if app_name in registry:
        local_nursery = registry.pop(app_name)
        local_nursery.cancel_scope.cancel()


async def _app_scope(app: app_spec, task_status=trio.TASK_STATUS_IGNORED):
    if app.permanent:
        restart = supervisor.restart_strategy.PERMANENT

    else:
        restart = supervisor.restart_strategy.TRANSIENT

    async with trio.open_nursery() as nursery:
        task_status.started(nursery)

        children = [
            supervisor.child_spec(
                id=app.module.__name__,
                task=app.module.start,
                args=[app.start_arg],
                restart=restart,
            )
        ]
        opts = app.opts if app.opts is not None else supervisor.options()

        nursery.start_soon(supervisor.start, children, opts)



import inspect
import sys

from types import ModuleType


def current_module() -> ModuleType:
    """
    This function should be called at the root of a module.

    :returns: The current module (similar to `__name__` for the current module name)

    .. code-block:: python
       :caption: Example

       from triotp.helpers import current_module

       __module__ = current_module()  # THIS WORKS


       def get_module():
           return current_module()  # THIS WON'T WORK
    """

    stack_frame = inspect.currentframe()

    while stack_frame:
        if stack_frame.f_code.co_name == "<module>":
            if stack_frame.f_code.co_filename != "<stdin>":
                caller_module = inspect.getmodule(stack_frame)

            else:
                caller_module = sys.modules["__main__"]

            if caller_module is not None:
                return caller_module

            break

        stack_frame = stack_frame.f_back



"""
A dynamic supervisor is almost identical to a normal supervisor.

The only difference is that a dynamic supervisor creates a mailbox in order to
receive requests to start new children from other tasks.

.. code-block:: python
   :caption: Example

   # app.py

   from triotp import supervisor, dynamic_supervisor
   import trio

   from . import worker


   async def start():
       opts = supervisor.options()
       children = [
           supervisor.child_spec(
               id='worker_pool',
               task=dynamic_supervisor.start,
               args=[opts, 'worker-pool'],
           ),
       ]

       async with trio.open_nursery() as nursery:
           await nursery.start_soon(supervisor.start, children, opts)

           await dynamic_supervisor.start_child(
               'worker-pool',
               supervisor.child_spec(
                   id='worker-0',
                   task=worker.start,
                   args=[],
                   restart=supervisor.restart_strategy.TRANSIENT,
               ),
           )
"""

from triotp import supervisor, mailbox
import trio

from typing import Optional, Union


async def start(
    opts: supervisor.options,
    name: Optional[str] = None,
    task_status=trio.TASK_STATUS_IGNORED,
) -> None:
    """
    Starts a new dynamic supervisor.

    This function creates a new mailbox to receive request for new children.

    :param opts: Supervisor options
    :param name: Optional name to use to register the supervisor's mailbox
    :param task_status: Used to notify the trio nursery that the supervisor is ready
    :raises triotp.mailbox.NameAlreadyExist: If the `name` was already registered

    .. code-block:: python
       :caption: Example

       from triotp import dynamic_supervisor, supervisor
       import trio


       async def example():
           opts = supervisor.options()
           child_spec = # ...

           async with trio.open_nursery() as nursery:
               mid = await nursery.start(dynamic_supervisor.start, opts)
               await dynamic_supervisor.start_child(mid, child_spec)
    """

    async with mailbox.open(name) as mid:
        task_status.started(mid)

        async with trio.open_nursery() as nursery:
            await nursery.start(_child_listener, mid, opts, nursery)


async def start_child(
    name_or_mid: Union[str, mailbox.MailboxID],
    child_spec: supervisor.child_spec,
) -> None:
    """
    Start a new task in the specified supervisor.

    :param name_or_mid: Dynamic supervisor's mailbox identifier
    :param child_spec: Child specification to start
    """

    await mailbox.send(name_or_mid, child_spec)


async def _child_listener(
    mid: mailbox.MailboxID,
    opts: supervisor.options,
    nursery: trio.Nursery,
    task_status=trio.TASK_STATUS_IGNORED,
) -> None:
    task_status.started(None)

    while True:
        request = await mailbox.receive(mid)

        match request:
            case supervisor.child_spec() as spec:
                await nursery.start(supervisor._child_monitor, spec, opts)

            case _:
                pass



"""
TriOTP logging system relies on the Logbook_ library. Each node has its own log
handler.

.. _logbook: https://logbook.readthedocs.io/
"""

from enum import Enum, auto
import logbook


class LogLevel(Enum):
    """
    TriOTP node's logging level
    """

    NONE = auto()  #: Logging is disabled
    DEBUG = auto()
    INFO = auto()
    WARNING = auto()
    ERROR = auto()
    CRITICAL = auto()

    def to_logbook(self) -> int:
        """
        Convert this enum to a Logbook log level.

        :returns: Logbook log level
        """

        return logbook.lookup_level(self.name)


def getLogger(name: str) -> logbook.Logger:
    """
    Get a logger by name.

    :param name: Name of the logger
    :returns: Logbook Logger instance
    """

    return logbook.Logger(name)



"""
TriOTP is built on top of the Trio_ async library. Therefore, it is not directly
compatible with AsyncIO libraries.

.. _trio: https://trio.readthedocs.io

This library revolves around the folllwing concepts:

 - a Node represent a single asynchronous loop (`trio.run`)
 - an application represent the root of a supervision tree
 - a supervisor handles automatic restart of child processes
 - a mailbox enables message passing between asynchronous tasks

On top of this concepts, this library provides:

 - generic servers to handle requests from other tasks
 - dynamic supervisors to schedule new tasks

.. _zeromq: https://zeromq.org/languages/python/

   **NB:** You don't get distributed computing out of the box like you would
   with Erlang_/Elixir_, this library is single-threaded and works within a
   Python application only.

.. _erlang: https://erlang.org
.. _elixir: https://elixir-lang.org/
"""
